# -*- coding: utf-8 -*-
"""
Here we have HTTP class for sending all requests
"""
from typing import Union, Callable
from typing import List, Tuple
from typing import Dict, Type
from typing import Any, Optional
from requests import get
from requests import Response
from urllib.parse import urlparse


class HTTP:
	'''
	This class from where we will make all requests

	example:
		HTTP.MakeObject() and this will return a requests class object without proxy

		:param Optional[str] proxy:
			(optional) Pass the proxy url
      
			example: 
				``socks5://admin:admin@127.0.0.1:9050``,
				``http://admin:admin@127.0.0.1:8000``,
				``https://127.0.0.1:8000``

	'''
	def __init__(self,*args:Tuple[str],**kwargs: Dict[str,Any]) -> None:
		self.args = args
		self.kwargs = kwargs

	def _GET(self, url:str, **kwargs: Dict[str,Any]) -> Response:
		self.kwargs.update(kwargs)
		return get((url), **self.kwargs)

	@classmethod
	def MakeObject(cls, proxy:Optional[str]=None) -> "HTTP":
		if proxy:
			u = urlparse(proxy)
			proxies = dict(
				http = f"{u.scheme}://{u.netloc}",
				https = f"{u.scheme}://{u.netloc}")
			return cls(proxies=proxies)
		return cls()



# -*- coding: utf-8 -*-
"""
Youtube.py is really lite python liberary with ultra fast downloader
"""

from .Config import (__title__,__version__,__issues__,
	__author__,__license__, __copyright__, __github__)
from .__main__ import Video, PlayList, HIGH, LOW, MID, Search
from .Connection import HTTP
from .Stream import Stream
from .Stream import Queue
from .Stream import FFMPEG
from .Cipher import Cipher
from .Error import *
from .Downloader import Download
import logging, sys 


# logging.basicConfig(level=logging.DEBUG, filename="mtube.log", filemode="w",
# 	format="%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s")





import logging
import argparse
import re
from youtube import Video, PlayList, __version__, __github__

class main:
	'''
	This is main cli class for youtube.py
	'''
	def __init__(self):
		self.args = self.process_args()

		self.url = self.check_url(self.args.url)

		if "playlist" in self.url:
			self.PlayList = True
			self.videos = [f"https://youtu.be/{n['vid']}" for n in PlayList(self.url).get_dict]
		else:
			self.PlayList = False
			self.videos = list([self.url])

		if self.args.output:
			self.output = self.args.output
		else:
			self.output = ""

		if self.args.logs:
			self.process_logg()

		if self.args.streams:
			self.print_streams()

		if self.args.ffmpeg:
			self.process_ffmpeg()

		if self.args.video:
			self.process_video()
		
		if self.args.itag:
			self.process_itag()

		if self.args.audio:
			self.process_audio()

		if self.args.resolution:
			self.process_resolution()
		exit()

	def print_streams(self):
		if self.args.streams:
			for n in self.videos:
				print(f'Video url: {n}')
				for m in Video(n).streams:
					print(m)

	def check_url(self, url:str) -> str:
		'''
		This will check the url type
		'''
		match = [re.compile(r'(http|https)://youtu.be/(.+)'),
				re.compile(r'(http|https)://(www.)?youtube.com/watch\?v=(.+)'),
				re.compile(r'(http|https)://(www.)?youtube.com/playlist\?list=(.+)'),]
		for n in match:
			if n.match(url):
				return url


	def process_ffmpeg(self):
		quality = self.parse_quality(self.args.ffmpeg)
		for n in self.videos:
			if self.PlayList:
				print(f'Downloading: {n}')

			if quality == 0:
				name = Video(n).streams.ffhigh.download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")

			elif quality == 1:
				name = Video(n).streams.ffmid.download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")
			
			else:
				name = Video(n).streams.fflow.download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")


	def process_video(self):
		quality = self.parse_quality(self.args.video)
		for n in self.videos:
			if self.PlayList:
				print(f'Downloading: {n}')

			if quality == 0:
				name = Video(n).streams.get_both.high().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")

			elif quality == 1:
				name = Video(n).streams.get_both.mid().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")
			
			else:
				name = Video(n).streams.get_both.low().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")


	def process_audio(self):
		quality = self.parse_quality(self.args.audio)
		for n in self.videos:
			if self.PlayList:
				print(f'Downloading: {n}')

			if quality == 0:
				name = Video(n).streams.get_audios.high().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")

			elif quality == 1:
				name = Video(n).streams.get_audios.mid().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")
			
			else:
				name = Video(n).streams.get_audios.low().download(dire=self.output, status=True)
				if name:
					print(f"{name} Downloaded")


	def process_resolution(self):
		resolution = self.args.resolution
		for n in self.videos:
			if self.PlayList:
				print(f'Downloading: {n}')

			streams = Video(n).streams
			name = streams.ffresolution(resolution).download(dire=self.output, status=True)
			if name:
				print(f"{name} Downloaded")


	def process_itag(self):
		itag = self.args.itag
		for n in self.videos:
			if self.PlayList:
				print(f'Downloading: {n}')

			name = Video(n).streams.get_itag(itag).download(dire=self.output, status=True)
			if name:
				print(f"{name} Downloaded")


	def parse_quality(self,q:str) -> int:
		if q.lower() == "high":
			return 0
		elif q.lower() == "mid":
			return 1
		else:
			return 2


	def process_logg(self):
		logging.basicConfig(level=logging.DEBUG,
			format="%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s")


	def process_args(self):
		parser = argparse.ArgumentParser(description="Definitions:\n  Quality Types: ('HIGH','MID', 'LOW')\n"
			"  Video resolution examples. ('144', '240', '360', '480', '720', '1080') more/less depend on your video\n"
			"\nNote:\n  Default values for '--ffmpeg', '--video', '--audio' is 'HIGH'\n"
			"  Default values for '--resolution' is 720",
											formatter_class=argparse.RawTextHelpFormatter)
		parser.add_argument('url', help='Enter video or playlist url')
		parser.add_argument('-o', '--output', type=str, help='Output location for downloading streams.')
		parser.add_argument('-l', '--logs', action='store_true', help='To enable extra logs.')

		group = parser.add_mutually_exclusive_group()

		group.add_argument('-ff', '--ffmpeg', const="HIGH", nargs="?", type=str, help='FFMPEG downloads audio/video both then copy in one file, Pass the quality type.')
		group.add_argument('-v', '--video', const="HIGH", nargs="?", type=str, help='Downlaod video (only progressive), Pass quality.')
		group.add_argument('-i', '--itag', type=int, help='Download stream with itag, Pass itag id.')
		group.add_argument('-s', '--streams', action='store_true', help='This argument will list the all available streams.')
		group.add_argument('-a', '--audio', const="HIGH", nargs="?", type=str, help='Download audio, Pass quality.')
		group.add_argument('-r', '--resolution', const=720, nargs="?", type=int, help='Download stream with quality type, Pass video resolution.')
		group.add_argument('--version', action="version", version=f'Youtube.py version {__version__} ({__github__})',help='Check the current version of youtube.py')
		args = parser.parse_args()
		return args


if __name__ == '__main__':
	main()


# -*- coding: utf-8 -*-
"""
Here we have cipher class for deciphering 
youtube signature

"""
from typing import Union, Callable
from typing import List, Tuple 
from typing import Dict, NoReturn
from typing import Any, Optional
from re import match, findall
import logging
from .Error import RegexError
from .Config import (__issues__,__github__)


logg = logging.getLogger(__name__)

def Reverse(arr:List[str], _:int) -> List[str]:
	'''
	here we will reverse the string

	example:
	>>> Reverse(['n','1','3'])
	['3', '1', 'n']
	'''
	return arr[::-1]

def Splice(arr:List[str], b:int) -> List[str]:
	'''
	here we will reverse the string

	example:
	>>> Splice(['n','1','3'],1)
	['1','n']
	'''
	return arr[b:]

def Switch(arr:List[str], b:int) -> List[str]:
	'''
	here we will swap the some chars

	var c = a[0];a[0] = a[b % a.length];a[b] = c
	'''
	tmp=arr[0]
	arr[0]=arr[b%len(arr)]
	arr[b]=tmp
	return arr

class Cipher():
	'''
	Here we will decrypt the signature of youtube videos

	js is the raw string javascript of base.js

	:param str js:
		pass javascript of base.js

	:rtype: None
	:returns: nothing
	'''
	def __init__(self,js:str) -> None:
		'''
		js is the raw string javascript of base.js

		:param str js:
			pass javascript of base.js

		:rtype: None
		:returns: nothing
		'''
		self.funs_methods = self.Get_funs(js)
		variable_name = self.funs_methods[0].split(".")[0]
		self.js_func = self.Get_js_funcs(js,variable_name)
		self.convert_funcs = self.Get_funs_objects(self.js_func)

	def Get_signature(self,sig:List[str]) -> str:
		'''
		This function will return decipher signarure

		:param list[str] sig:
			cipher signature provied by youtube

		:rtype: str
		:returns: will return decipher signature
		'''
		sig=list(sig)
		for n in self.funs_methods:
			func_name, b = self.Parsec_js_methods(n)
			sig=self.convert_funcs[func_name](sig,b)
		return "".join(sig)


	@staticmethod
	def Parsec_js_methods(s:str) -> Tuple[str, int]:
		'''
		This the parse the methods

		Example:

		>>> Parsec_js_methods("wv.TY(a,37)")
		('TY',37)
		'''
		d=s.split(".")[1]
		return d.split("(")[0], int(d.split(",")[1][:-1])
	@staticmethod
	def Func_check(s:str) -> Callable[[List[str],int], List[str]]:
		'''
		This will check compair the js function
		and return the python object

		:param str s:
			pass the js fuction

		:rtype: object
		:returns: will return python object

		example:

		>>> Func_check('TY:function(a){a.reverse()')
		<function Reverse at 0x7fd281937710>
		'''

		check = [
			{"c":["reverse"], "func": Reverse},
			{"c":["splice"], "func" : Splice},
			{"c":[ ".length","%",], "func" : Switch},
		]
		
		for n in check:
			for m in n["c"]:
				if m in s:
					return n["func"]


	def Get_funs_objects(self,funs:List[str]) -> Dict[str, Callable[[List[str],int], List[str]]]:
		'''
		This function will return the dict with
		javascript function name and object

		:param List[str] funs:
			pass the list of function

		:rtype: Dict[str, Callable[[List[str],int], List[str]]]
		:returns: dict of function name and function object

		example:

		>>> Get_funs_objects(['TY:function(a){a.reverse()}', 'A3:function(a,b){a.splice(0,b)}',
			'vd:function(a,b){var c=a[0];a[0]=a[b%a.length];a[b%a.length]=c}}'])
		{'TY': <function Reverse at 0x7f0f9cfce7a0>, 'A3': <function Splice at 0x7f0f9cfd63b0>, 
			'vd': <function Switch at 0x7f0f9cfd6710>}
		'''
		f={}
		for n in funs:
			tm=n.split(":")
			f[tm[0]] = self.Func_check(tm[1])
		return f

	@staticmethod
	def Get_funs(js:str) -> List[str]:
		'''
		Will return the method of encryption

		:param str js:
			pass the raw js of base.js

		:rtype: List[str]
		:returns: the decipher method/algorithm in list of strings

		example:

		>>> Get_funs(js)
		... ['wv.TY(a,37)', 'wv.vd(a,10)', 'wv.TY(a,7)', 'wv.A3(a,1)', 'wv.vd(a,13)', 'wv.A3(a,2)', 'wv.vd(a,15)']
		'''
		match=r'function\(a\){a=a.split\(""\);(.*);return a.join\(""\)};'
		d=findall(match,js)[0].split(";")
		if d:
			logg.debug(f"Found algorithm for decipher in JS")
			return d
		else:
			logg.error("We do not ")
			raise RegexError(f"{match} Pattern do not found in base.js, please contact {__issues__}, help us make it more stable for everyone")
	@staticmethod
	def Get_js_funcs(js:str,var:str) -> List[str]:
		'''
		will return the js raw function as string

		:param str js:
			raw js html
		:param str var:
			js variable name

		:rtype: List[str]
		:returns: the variable content/program from base.js

		Example.

		>>> Get_js_funcs(js, var)
		... ['TY:function(a){a.reverse()', 'A3:function(a,b){a.splice(0,b)', 
			'vd:function(a,b){var c=a[0];a[0]=a[b%a.length];a[b%a.length]=c}']
		'''
		match=r"var "+var+"={(.*?)};"
		d=findall(match,js.replace("\n",""))[0].split("},")
		if d:
			logg.debug(f"Variable: <{var}> found in base.js for decipher the signature")
			return [n+"}" for n in d]
		else:
			logg.error(f"Variable <{var}> do not found in base.js, please contact {__issues__}")
			raise RegexError(f"Variable <{var}> do not found in base.js, please contact {__issues__}")


# -*- coding: utf-8 -*-

__author__ = 'Mayank Gupta'
__version__ = '1.1'
__license__ = 'License :: MIT License'

from typing import Union, Callable
from typing import List, Tuple
from typing import Dict, Type
from typing import Any, Optional
import socket,select,re,ssl,threading,sys,os
from urllib.parse import urlparse, unquote
from time import sleep, time
import tempfile, os, logging
from select import select
from random import randint

logg = logging.getLogger(__name__)

class Download(object):
	'''
	This :class:`Download <Download>` will download streams with multi-connections.

	:param str url: 
		Pass the download link
	:param str name: 
		(optional) Pass the name of file name. 
	:param str dire: 
		(optional) Pass the dir for output file with excluding filename.
	:param bool status: 
		(optional) Pass if you want to enable process bar. **Default[False]**
	:param int connection: 
		(optional) Pass number is connection to be create. **Default[8]**
	:param int chunk: 
		(optional) Pass the chunk/buffer size for accepting packets. **Default[5120]**

	:rtype: str
	:returns: the file name

	'''
	def __init__(self,url:str,dire:str="",name:str="",status:bool=False,connection:int=8, chunk:int = 5120) -> None:
		self.name = name.replace(" ","_")
		self.dire = dire
		if self.dire:
			self.dire = self.create_user_dir(dire)
			if self.dire[-1] == "/":
				self.dire = self.dire[:-1]
		self.status = status
		self.chunk = chunk
		self.connection = connection
		self.url = unquote(url)

	def start(self) -> str:
		'''
		Start will fire up the downloading
		'''
		protocol, url, host =self.RawData(self.url)
		logg.debug(f'protocol: {protocol}, host: {host}')

		check = self.check_multi(protocol, self.url, host)
		logg.debug(f'Download check status: {check}')
		if check[0] == "0":

			# Data for process bar
			# gg= bytes downloaded, size total size of file,
			# when is bool vaule to start and stop the process bar
			self.size = int(self.header["content-length"])

			self.gg = 0 # download bytes
			self.when = True

			#get filename
			name = self.getfilename()

			logg.debug(f'Filename: {name}, Filesize: {self.size}')

			# Create ranges for downloading chunks in parts
			ranges = self.get_range(int(self.header["content-length"]),self.connection)

			self.files = {}
			threads = []
			for n,m in enumerate(ranges):
				req=self.gen_req(host,url,{"range":f"bytes={m}"})
				threads.append(threading.Thread(target=self.down, args=(protocol, host, req, m, str(n))))
				# break

			if self.status:threading.Thread(target=self.run).start()
			for n in threads:n.start()
			for n in threads:n.join()
			
			# End of process bar 
			self.when = False

			with open(name,"wb") as f:
				for n in range(len(self.files)):
					ff=self.files[n]
					ff.seek(0)
					f.write(ff.read())
					ff.close()
			f.close()

			# end of procedd bar with 100%
			p=int(int(self.gg)*50/int(self.size))
			if self.status:print("Process: [{}] {}% Complete {:<10}".format("█"*p+"-"*(50-p), p*100/50,"0.0 Kb/s"))
			logg.debug(f"Downloading conpleted 100% Filename{name}")

			# print(name)
			return name

		elif check[0] == "1" :
			name = self.getfilename()
			req=self.gen_req(host,url)
			sock=self.connect(protocol,host)
			sock.sendall(req)
			data=sock.recv(self.chunk)
			header,image=self.hparsec(data)
			f = open(name,"wb")
			f.write(image)


			# gg= bytes downloaded, size total size of file,
			# when is bool vaule to start and stop the process bar
			self.gg = len(image)
			self.size = int(header["content-length"]) 
			self.when = True

			#Start The process bar if status TRUE
			if self.status:threading.Thread(target=self.run).start()

			logg.debug(f'Filename: {name}, Filesize: {self.size}')

			while True:
				try:
					data = sock.recv(self.chunk)
					if not data:break
					f.write(data)
					self.gg += len(data)
				except socket.timeout:
					break

			#End od process bar
			self.when = False

			# end of procedd bar with 100%
			p=int(int(self.gg)*50/int(self.size))
			if self.status:print("Process: [{}] {}% Complete {:<10}".format("█"*p+"-"*(50-p), p*100/50,"0.0 Kb/s"))

			# Return the file name
			return name

		elif check[0] == "2" :
			name = self.getfilename()
			req=self.gen_req(host,url)
			sock=self.connect(protocol,host)
			sock.sendall(req)
			data=sock.recv(self.chunk)
			header,image=self.hparsec(data)
			f = open(name,"wb")
			f.write(image)

			if self.status:
				logg.debug("We can't run status bar for this, No content-length found")

			logg.debug(f'Filename: {name}, Filesize: Unknown')

			while True:
				try:
					data = sock.recv(self.chunk)
					if not data:break
					f.write(data)
				except socket.timeout:
					break
					
			# Return the file name
			return name
		else:
			return check[1]

	def create_user_dir(self,foldername:str) -> str:
		if not os.path.exists(foldername):
			os.makedirs(foldername)
		return foldername

	def rangediff(self,s):
		c,b = s.split("-")
		c,b = int(c),int(b)
		if self.size == b:
			diff = b-c
			return diff
		else:
			diff = b-c
			return diff+1

	def down(self, protocol:str, host:str, req:bytes, range:list, id:str="") -> None:
		f = tempfile.TemporaryFile()
		if id != "":self.files[int(id)] = f
		sock=self.connect(protocol,host)
		diff = self.rangediff(range)
		sock.settimeout(5)
		sock.sendall(req)
		data=sock.recv(self.chunk)
		header,image=self.hparsec(data)
		self.gg += len(image)
		local_gg = 0
		local_gg =+len(image)
		f.write(image)
		while True:
			try:
				data = sock.recv(self.chunk)
				if not data:break
				f.write(data)
				self.gg += len(data)
				local_gg =+len(data)
				if local_gg >= diff: break
			except socket.timeout:
				break

		f.seek(0)

	def run(self):
		self.temp1=0
		while self.when:
			speed=(self.gg-self.temp1)/1024
			p=int(int(self.gg)*50/int(self.size))
			print("Process: [{}] {}% Complete {:<8}Kb/s".format("█"*p+"-"*(50-p), p*100/50,"{:.2f}".format(speed)),end="\r")
			self.temp1=self.gg
			sleep(1)

	def get_range(self, length:int, conn:int) -> List[str]:
		av = int(length/conn)
		r=[]
		start = 0
		r.append(f'{start}-{start+av}')
		start+=av
		if conn>1:
			for n in range(conn-2):
				r.append(f'{start+1}-{start+av}')
				start+=av
			r.append(f'{start+1}-{length}')
		return r

	def getfilename(self) -> str:
		finalname = ""
		name = ""
		if self.dire:
			if not self.name:
				if self.tmpname:
					finalname = f'{self.dire}/{randint(10,99)}{self.tmpname}'
				else:
					dd=self.header["content-type"].split("/")[1].split("+")[0]
					finalname = f'{self.dire}/{randint(10,99)}{int(time())}.{dd}'
			else:finalname = f'{self.dire}/{randint(10,99)}{self.name}'
		else:
			if not self.name:
				if self.tmpname:
					finalname = f'{randint(10,99)}{self.tmpname}'
				else:
					dd=self.header["content-type"].split("/")[1].split("+")[0]
					finalname = f'{randint(10,99)}{int(time())}.{dd}'
			else:finalname = f'{randint(10,99)}{self.name}'

		for n in finalname:
			if n not in '\\ /:*?"<>|':
				name+=n
				
		return name

	def check_multi(self, protocol:str, url:str, host:str) -> Tuple:
		req=self.gen_req(host,url)
		sock=self.connect(protocol,host)
		sock.sendall(req)
		data=sock.recv(self.chunk)
		self.header,image=self.hparsec(data)
		if "content-length" in self.header.keys():
			if int(self.header["status"]) != 200:
				try:
					sock.close()
					name = self._Download(self.header["location"], dire=self.dire, name=self.name, status=self.status, chunk=self.chunk, connection=self.connection)
					return "2",name
				except Exception as err:
					print(f"Error: {err}")
					print("We cant download from this URL Contact Admin with URL OR can't save with this file name")
					sock.close()
					sys.exit(1)

		else: return "2",""

		if "accept-ranges" in self.header.keys():
			return "0",""
		return "1",""
	
	@classmethod
	def _Download(cls,*args,**kwargs):
		return cls(*args,**kwargs).start()


	def connect(self, protocol:str, host:str) -> socket.socket:
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		if protocol=="https":
			s.connect((host, 443))
			s = ssl.create_default_context().wrap_socket(s, server_hostname=host)
		elif protocol=="http":
			s.connect((host, 80))
		else:
			print("we only support HTTP and HTTPS")
			s.close()
			sys.exit(1)
		return s

	def hparsec(self,data:bytes) -> Tuple[Dict[str,str], bytes]:
		header =  data.split(b'\r\n\r\n')[0]
		store =  data[len(header)+4:]
		html = data[len(header)+4:]
		header=header.decode().split("\r\n")

		out={}
		for n in header[1:]:
			temp=n.split(":")
			value=""
			for n in temp[1:]:
				value+=n+":"
			out[temp[0].lower()]=value[1:len(value)-1]
		out["status"]=header[0].split()[1]

		return out,store

	def gen_req(self, host:str, url:str, header:Dict[str,str] = {}) -> bytes:
		req=f'GET {url} HTTP/1.1\r\nhost: {host}\r\nuser-agent: MayankFawkes/bot\r\nconnection: close\r\n'
		for n, m in header.items():
			req += f'{n}:{m}\r\n'
		req+="\r\n"
		return req.encode()

	def RawData(self,web_url:str)-> Tuple[str, str, str]:
		o=urlparse(web_url)
		host=o.netloc
		protocol=o.scheme
		if o.query:
			url=(o.path+"?"+o.query)
			self.tmpname = ""
		else:
			url=o.path
			self.tmpname = o.path.split("/")[-1]
		return protocol, url, host

if __name__ == '__main__':
	# link=input("Enter Url -->")
	link='https://storge.pic2.me/download/origin/257714.jpeg'
	# link="https://r2---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606610611&ei=U5rCX9fZKImlwgPP0KOwDQ&ip=171.60.177.128&id=o-AOTEor8QB7nEzOV58ZY1GegCDzqGoCJFH4yqW6Scpeqh&itag=248&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=DC&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvh6&ms=au%2Crdu&mv=m&mvi=2&pcm2cms=yes&pl=20&gcr=in&initcwndbps=695000&vprv=1&mime=video%2Fwebm&ns=vUXVdwbvoJzjIECjxScz6IUF&gir=yes&clen=12295247&dur=193.680&lmt=1606586572289453&mt=1606588849&fvip=2&keepalive=yes&c=WEB&txp=5532432&n=y9n0dSPriQCxsQyupR2&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cgcr%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpcm2cms%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAPE6C_TL49N8jIqMp9hVgj9bqs8zY93wOjjW7ZONuFo2AiEA81M8BnT1W519Y90XZhYpaF0wmSMvFMSKNGhmggOcPdE%3D&sig=AOq0QJ8wRgIhAOAdoYf8jUSErCo7XlxjSb4ePQIYpS2IRzUaIW5_Fk0GAiEAgbqa68gBBkWUdml0tzts_dphR7JxQXSQ8C5AhQnV86s="
	# link="http://www.macaronisoup.com/songs/mp3/LoobyLoo.mp3"
	# link = "https://portswigger.net/burp/releases/download?product=community&version=2020.11.2&type=WindowsX64"
	dd=Download(link ,name = "", status = True, connection = 8, chunk = 5120).start()
	print(dd)


# -*- coding: utf-8 -*-
"""
Here are some function which helps extracting info
online information <GET requests>

"""
from typing import Union, Callable
from typing import List, Tuple 
from typing import Dict
from typing import Any
from .Connection import HTTP
from re import compile, findall
from urllib.parse import unquote
import json, logging
from .Error import (RegexError, VideoInfo, GeoBlockingError, ParsecError)
from .Config import (__issues__,__github__)

logg = logging.getLogger(__name__)

def Get_js(req:HTTP,vid:str) -> Tuple[str, str]:
	'''
	Thie will return the url of base.js
	
	Example..
	>>> Get_Js("M3mJkSqZbX4")
	'https://www.youtube.com/s/player/XXXXXXXXX/player_ias.vflset/en_GB/base.js'
	'''
	data=req._GET(f'https://www.youtube.com/watch?v={vid}').text
	try:
		match=compile(r'<script  src="(.*?)" type="text/javascript" name="player_ias/base" ></script>')
		src=findall(match,data)[0]
	except:
		match=compile(r'jsUrl":"(.*?)"')
		src=findall(match,data)[0]
	if src:
		return "https://www.youtube.com"+src, data
		logg.debug("Successfully got raw js <player_ias/base.js>")
	else:
		logg.error(f"Error raw js <player_ias/base.js> not found <VID:{vid}>")
		raise RegexError(f"base.js not found please report here {__issues__} with {vid}, help us make it more stable for everyone")

def Is_age_restricted(rawhtml:str) -> bool:
	'''
	This will return true or False

	>>> is_age_restricted(htmldata)
	True/False


	'''
	if "og:restrictions:age" in rawhtml:
		return True
	return False

def Get_video_info(req:HTTP,url:str) -> Tuple[Any, bool, Dict[str, str]]:
	'''
	This willl return video info 
	from: https://youtube.com/get_video_info

	:args str url
		url in string <get_video_info>

	:rtype: Tuple[Any, bool, Dict[str, str]]
	:returns: the extracted data, true if signature found, raw data dict
	'''
	try:
		r=req._GET(url).text
		Fdata = {unquote(o.split("=")[0]):unquote(o.split("=")[1]) for o in r.split("&")}
		data = json.loads(Fdata["player_response"])
		r = data = json.loads(Fdata["player_response"]) # to return this  
		assert CheckPlayable(data["playabilityStatus"]) == True, f'Video is {data["playabilityStatus"]["status"]}, Please use Proxy/VPN to bypass Geo-blocking'
		data = GetStreamingData(data["streamingData"])
		# data = data["streamingData"]["adaptiveFormats"] + data["streamingData"]["formats"]
		issig:bool = Videodata_parse(data)
		logg.debug(f"We Successfully received video data/info")
		return data, issig, r
	except:
		logg.error(f"Video Info not found <{url}> please contact <{__issues__}>")
		raise VideoInfo(f"Video Info not found <{url}> please contact <{__issues__}>, help us make it more stable for everyone")
def Videodata_parse(data:Dict[Any, Any]) -> bool:
	'''
	Takes Data and parsec the signatureCipher
	and upadte the url and del the signatureCipher
	if it has or return the same
	'''
	logg.debug(f"Videodata processing ..... ")
	sig:bool=False
	for n in data:
		if "signatureCipher" in n.keys():
			sig=True
			url = _parsesignatureCipher(n["signatureCipher"])
			del n["signatureCipher"]
			n.update(url)
		else:
			sig=False
			continue
	logg.debug(f"Videodata processing Completed, Signature found: <{sig}> ")
	return sig

def _parsesignatureCipher(signatureCipherraw:str) -> Dict[Any, Any]:
	'''
	this will convert the str of signatureCipher
	to dict and return
	'''
	data={unquote(o.split("=")[0]):unquote(o.split("=")[1]) for o in signatureCipherraw.split("&")}
	return data


def Req_get(req:HTTP,url:str) -> str:
	'''
	Send simple get request with object
	'''
	return req._GET(url).text

def CheckPlayable(data:Dict[Any, Any]) -> bool:
	'''
	Here we will check is video is really
	playable or not/ Check if video is aviable
	for you or in your country.
	'''
	if data["status"] !="OK":
		logg.info(f'Video is {data["status"]}')
		raise GeoBlockingError(f'Video is {data["status"]}, {data["errorScreen"]["playerErrorMessageRenderer"]["subreason"]["runs"][0]["text"].replace("+"," ")}')
		return False
	return True


def GetStreamingData(data:Dict[Any, Any]) -> List[Any]:
	'''
	This will return all the streams with or without audio streams, ``adaptiveFormats`` and ``formats``.
	'''
	d=[]

	for n in data["formats"]:
		mtype, vcode=n["mimeType"].split(";+")
		n["mimeType"] = mtype
		mtype, forma = mtype.split("/")
		key, value = vcode.split("=")
		n["type"] = "both"
		n["format"] = forma
		n[key] = value.replace('"',"")
		d.append(n)

	for n in data["adaptiveFormats"]:
		mtype, vcode=n["mimeType"].split(";+")
		n["mimeType"] = mtype
		mtype, forma = mtype.split("/")
		key, value = vcode.split("=")
		n["type"] = mtype
		n["format"] = forma
		n[key] = value.replace('"',"")
		d.append(n)

	return d


def Get_playlist_videos(http:HTTP, url:str) -> Dict[str,str]:
	'''
	This will parse the initial json dict from web page source.

	:param HTTP http:
		 Pass the :class:`HTTP <HTTP>` object.
	:param str url:
		pass the plsylist url

	:rtype: Dict[str,str]
	:returns: The dict of of video details 
	'''
	html = http._GET(url).text.replace("\n","")
	f = ""
	try:
		match = compile(r'window\["ytInitialData"\] = (.*?);')
		f = findall(match, html)[0]
	except:
		match = compile(r'var ytInitialData = (.*?);')
		f = findall(match, html)[0]


	if f:
		return json.loads(f)
	else:
		logg.debug(f'RegexError we cannot found {match}')
		raise RegexError(f'RegexError we cannot found {match}')
	

def Continuation(http:HTTP,videos:List[Dict[str,str]], token:str) -> str:
	'''
	This will extract the data we will get from **browse_ajax** API.

	:param HTTP http:
		Pass the :class:`HTTP <HTTP>` object.
	:param List[Dict[str,str]] videos:
		Pass the videos dicts list.
	:param str token:
		Pass the token if any.

	:rtype: str
	:returns: The new/next continuous token if available.
	'''
	params = {
		"ctoken": token
	}
	headers={
		'X-YouTube-Client-Name': '1',
		'X-YouTube-Client-Version': '2.20200720.00.02'
	}
	token = ""
	data = http._GET("https://www.youtube.com/browse_ajax",params=params,headers=headers).json()[-1]
	for n in data["response"]["onResponseReceivedActions"][0]["appendContinuationItemsAction"]["continuationItems"]:
		if "playlistVideoRenderer" in n.keys():
			tmp = dict()
			tmp["vid"] = n["playlistVideoRenderer"]["videoId"]
			tmp["name"] = n["playlistVideoRenderer"]["title"]["runs"][0]["text"]
			videos.append(tmp)
		elif "continuationItemRenderer" in n.keys():
			token = n["continuationItemRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
		else:
			raise ParsecError(f'Serious Bug please report here {__issues__}')

	return token

def Parse_search_json(data:Dict[str,str], videos:List[Dict[str,str]]) -> bool:
	# print(json.dumps(data))
	for n in data["contents"]["twoColumnSearchResultsRenderer"]["primaryContents"]["sectionListRenderer"]["contents"][0]["itemSectionRenderer"]["contents"]:
		try:
			det = n["videoRenderer"]
			videos.append(dict(vid=det["videoId"],name=det["title"]["runs"][0]["text"]))
		except:
			continue
	return True



def Search_results(http:HTTP, query:str, videos:List[Dict[str,str]],country:str) -> bool:
	'''
	This will parse the initial json dict from web page source.
	
	:param HTTP http:
		Pass the :class:`HTTP <HTTP>` object.
	:param str query:
		pass query for searching.
	:param str videos:
		pass the video list of dicts
	:param str country:
		pass the country code

	:rtype: bool
	:returns: true or false depend on search results.

	'''
	SEARCH_URL = "https://www.youtube.com/results"
	PARAMS = {
		"search_query":query,
		"gl":country.upper()
	}
	HEADERS = {}

	html = http._GET(SEARCH_URL, params=PARAMS,headers=HEADERS).text.replace("\n","")
	try:
		match = compile(r'var ytInitialData = (.*?);')
		f = findall(match, html)[0]
		f = json.loads(f)
	except:
		logg.debug(f'RegexError we cannot found {match}')
		raise RegexError(f'RegexError we cannot found {match}')

	if f:
		if Parse_search_json(f,videos):
			return True
		else:
			return False



# -*- coding: utf-8 -*-
"""
This file contain all the error classes
"""

class MtubeError(Exception):
	pass

class ParsecError(Exception):
	pass

class RegexError(Exception):
	pass

class VideoInfo(Exception):
	pass

class InvalidVideoUrl(Exception):
	pass

class GeoBlockingError(Exception):
	pass

class FFMPEG_Not_Found(Exception):
	pass

class ResolutionError(Exception):
	pass


# -*- coding: utf-8 -*-
"""
Here we have 2 classes where we will
manage/sort the streams and return the 
download object
"""

from typing import Union, Callable
from typing import List, NoReturn
from typing import Dict, Optional
from typing import Any, Tuple
from subprocess import run
from .Error import (FFMPEG_Not_Found, ResolutionError)
from .Downloader import Download
from .Config import __docs__
import logging, time, os

logg = logging.getLogger(__name__)

class Stream:
	'''
	Here we will process :class:`Stream <Stream>`

	:param Dict[Any,Any] stream: singal dict of stream json
	:param str name: Pass the video title

	'''

	def __init__(self,stream:Dict[Any, Any], name:str) -> None:
		
		self.stream = stream
		self.name = name

		rep = f'<Class {self.__class__.__name__}: '
		if stream["type"] != "audio":
			rep += f'itag="{self.stream["itag"]}" '
			rep += f'qualityLabel="{self.stream["qualityLabel"]}" '
			rep += f'fps="{self.stream["fps"]}" '
			rep += f'mime="{self.stream["mimeType"]}" '
			rep += f'type="{self.stream["type"]}" '
			rep += f'format="{self.stream["format"]}" '
			rep += f'codecs="{self.stream["codecs"]}"'
			rep += f'>'
		else:
			rep += f'itag="{self.stream["itag"]}" '
			rep += f'bitrate="{self.stream["bitrate"]}" '
			rep += f'mime="{self.stream["mimeType"]}" '
			rep += f'type="{self.stream["type"]}" '
			rep += f'format="{self.stream["format"]}" '
			rep += f'codecs="{self.stream["codecs"]}"'
			rep += f'>'
		self.rep = rep

	def __str__(self) -> str:
		return self.rep

	def __repr__(self) -> str:
		return self.rep
	
	def check(self,a:str) -> str:
		'''
		Here we will take a and return the values,
		``a`` could be  ``qualityLabel`` or ``type``
		qualityLabel: will return ``resolution of the video``
		type: will return ``video, audio or both``
		'''
		if a == "qualityLabel":
			if self.stream["type"] != "audio":
				return int(self.stream[a].replace("p",""))
			else:
				return self.stream["bitrate"]
		if a == "type":
			return self.stream["type"]

	@property
	def is_audio(self) -> bool:
		'''
		Check is the stream is audio

		:rtype: bool
		:returns: check the stream and return bool(true, false)
		'''
		if self.stream["type"] == "audio":
			return True
		return False
	
	@property
	def is_video(self) -> bool:
		'''
		Check is the stream is video

		:rtype: bool
		:returns: check the stream and return bool(true, false)
		'''
		if self.stream["type"] == "video":
			return True
		return False

	@property
	def is_both(self) -> bool:
		'''
		Check is the stream has audio/video

		:rtype: bool
		:returns: check the stream and return bool(true, false)
		'''
		if self.stream["type"] == "both":
			return True
		return False
	
	@property
	def url(self) -> str:
		'''
		This will return the URL of thie Stream

		:rtype: str
		:returns: url of video in str
		'''
		return self.stream["url"]

	@property
	def format(self) -> str:
		'''
		This will return the stream format

		:rtype: str
		:returns: format of video in str
		'''
		return self.stream["format"]
	
	@property
	def itag(self) -> int:
		'''
		This will return the stream format

		:rtype: int
		:returns: itag of video in int
		'''
		return self.stream["itag"]
	
	def download(self, dire:str="", name:str="",status:bool=False,connection:int=8, chunk:int=5120) -> Download:
		'''
		This will return the :class:`Download <Download>`
		
		:param str name: 
			(optional) Pass the name of file name. 
		:param str dire: 
			(optional) Pass the dir for output file with excluding filename.
		:param bool status: 
			(optional) Pass if you want to enable process bar. ``Default[False]``
		:param int connection: 
			(optional) Pass number is connection to be create. ``Default[8]``
		:param int chunk: 
			(optional) Pass the chunk/buffer size for accepting packets. ``Default[5120]``

		:rtype: :class:`Download <Download>`
		:returns: :class:`Download <Download>` object
		'''
		fname = self.filename
		if name:
			fname=name
		return Download(url=self.stream["url"], dire=dire, name=fname , status=status, connection=connection, chunk=chunk).start()


	@property
	def filename(self) -> str:
		'''
		This will generate file name (this function is for internal use)

		:rtype: : str
		:returns: filename in str
		'''
		return f'{self.name}.{self.stream["format"]}'


class FFMPEG:
	'''
	Here we will download 2 stream audio/video, 
	and copy them in one file with ffmpeg

	:param List[Stream] av:
		pass the stream of audio and video in list,

	:rtype: None
	:returns: Nothing will get return
	'''
	def __init__(self,av:List[Stream]) -> None:
		if not self.check_ffmpeg():
			pass
		self.av = av
	def __repr__(self):
		return f"{self.av}"
	def download(self,dire:str="", name:str="",status:bool=False,connection:int=8, chunk:int=5120) -> str:
		'''
		This will return bool value as video download convert status`
		
		:param str name: 
			(optional) Pass the name of file name. 
		:param str dire: 
			(optional) Pass the dir for output file with excluding filename.
		:param bool status: 
			(optional) Pass if you want to enable process bar. **Default[False]**
		:param int connection: 
			(optional) Pass number is connection to be create. **Default[8]**
		:param int chunk: 
			(optional) Pass the chunk/buffer size for accepting packets. **Default[5120]**

		:rtype: str
		:returns: name of the file
		'''
		args = ['ffmpeg -i', '-i','-c copy -c:a aac']
		names = []
		for n in self.av:
			logg.debug(f"Downloading : {n.check('type')} : {n.format}: {n.check('qualityLabel')}")
			if status:print(f"Downloading : {n.check('type')} : {n.format}: {n.check('qualityLabel')}")
			nam = n.download(dire=dire, name=name , status=status, connection=connection, chunk=chunk)
			names.append(f'{nam}')
			logg.debug(f"{nam} Downloaded")
		tmp=nam.split("/")
		names.append(f'{"/".join([n for n in tmp[:-1]] + [tmp[-1][2:].split(".")[0]])}.mp4')
		cmd = [f'{n} {m}' for n,m in zip(args,names)]
		cmd = " ".join(cmd).split()

		logg.debug(f"Merge audio/video file in one file {args[-1]}")

		if self.runcmd(cmd=cmd):
			os.remove(names[0])
			os.remove(names[1])
			logg.debug(f"Video/audio downloaded and merged {args[-1]}")
			return names[-1]
		else:
			logg.error(f"Error while converting {args[0]} and {args[1]} to {args[-1]}")
			return ""



	def check_ffmpeg(self):
		'''
		Checking if FFMPEG installed or not

		:rtype: bool
		:returns: check status in (True/False)
		'''
		if "hyper fast audio and video" in self.runcmd(cmd="ffmpeg"):
			logg.debug(f"FFMPEG Found")
			return True
		return False

	def runcmd(self, cmd:str) -> str:
		'''
		Here we will run the code and return
		'''
		try:
			return str(run(cmd, capture_output=True, text=True, input='y\n')).lower()
		except FileNotFoundError:
			if cmd[0] == "ffmpeg":logg.debug(f"FFMPEG Not-Found")
			raise FFMPEG_Not_Found(f'Please install ffmpeg, check {__docs__}')
	



class Queue:
	'''
	We will manage the :class:`Stream <Stream>` queues here

	:param List[Stream] data:
		pass the dict of :class:`Stream <Stream>` objects.
	:param bool repeat:
		if you re generating Queue object from inside the function.
	
	'''
	def __init__(self,data:List[Stream], repeat:bool = False) -> None:
		self.All = sorted(data, key=lambda x: x.itag)
		self.data = data
		if not repeat:
			self.data = sorted(self.init(data),key=lambda x: x.itag)

	def __str__(self) -> str:
		return f"{self.data}"

	def __repr__(self) -> str:
		return f"{self.data}"

	def __getitem__(self, i: Union[int, slice]) -> Union[Stream, List[Stream]]:
		return self.data[i]

	@staticmethod
	def init(data:List[Stream]) -> List[Stream]:
		'''
		Here we will return the List[Stream]
		after removing 'ignore_itag'
		'''
		ignore_itag=[140]
		for n,m in enumerate(data):
			if m.itag in ignore_itag:
				del data[n]
		return data

	
	def format(self,f:str) -> "Queue":
		'''
		This will return all streams with
		format match

		:param str f:
			pass the format type ie. 'mp4'.

		:rtype: :class:`Queue <Queue>`
		:returns: :class:`Queue <Queue>` object


		'''
		return Queue(sorted([n for n in self.data if n.format == f], key=lambda x: x.check("qualityLabel")), repeat=True)

	def get_itag(self,i:int) -> Union[Stream, None]:
		'''
		This will return the stream,
		matched with i <arg>

		:param int i:
			pass the itag id for match.

		:rtype: Union[Stream, None]
		:returns: Will return :class:`Stream <Stream>` or None

		'''
		for n in self.data:
			if n.itag==i:
				return n

	@property
	def get_videos(self) -> "Queue":
		'''
		This will return all videos stream without audio.

		:rtype: :class:`Queue <Queue>`
		:returns: :class:`Queue <Queue>` object
		'''
		return Queue(sorted([n for n in self.data if n.is_video],key=lambda x: x.check("qualityLabel")), repeat=True)

	@property
	def get_audios(self) -> "Queue":
		'''
		This will return all audios stream without video.

		:rtype: :class:`Queue <Queue>`
		:returns: :class:`Queue <Queue>` object
		'''
		return Queue(sorted([n for n in self.data if n.is_audio],key=lambda x: x.check("qualityLabel")), repeat=True)

	@property
	def get_both(self) -> "Queue":
		'''
		This will return all audios/video stream.

		:rtype: :class:`Queue <Queue>`
		:returns: :class:`Queue <Queue>` object
		''' 
		return Queue(sorted([n for n in self.data if n.is_both],key=lambda x: x.check("qualityLabel")), repeat=True)


	def low(self,ff:bool=False) -> Stream:
		'''
		Low will return the lowest quality stream.

		:param bool ff:
			this input is not for the user please ignore.

		:rtype: :class:`Stream <Stream>`
		:returns: :class:`Stream <Stream>` object
		'''
		if ff:
			av = 0
			for n in range(av,av+2):
				if self.data[n].format == "webm":
					return self.data[n]
		else:return self.data[0]

	@property
	def fflow(self) -> FFMPEG:
		'''
		Processing with lowest quality video and audio

		:rtype: :class:`FFMPEG <FFMPEG>`
		:returns: :class:`FFMPEG <FFMPEG>` object
		'''
		vid = self.get_videos.low(ff=True)
		aud = self.get_audios.low(ff=True)
		return  FFMPEG([vid,aud])

	def high(self,ff:bool=False) -> Stream:
		'''
		High will return the highest quality stream.

		:param bool ff:
			this input is not for the user please ignore.

		:rtype: :class:`Stream <Stream>`
		:returns: :class:`Stream <Stream>` object
		''' 
		if ff:
			av = int(len(self.data))
			for n in range(av-2,av):
				if self.data[n].format == "webm":
					return self.data[n]
		else:return self.data[-1]
	
	@property
	def ffhigh(self) -> FFMPEG:
		'''
		Processing with highest quality video and audio

		:rtype: :class:`FFMPEG <FFMPEG>`
		:returns: :class:`FFMPEG <FFMPEG>` object
		'''
		vid = self.get_videos.high(ff=True)
		aud = self.get_audios.high(ff=True)
		return  FFMPEG([vid,aud])
	

	def mid(self,ff:bool=False) -> Stream:
		'''
		Mid will return the normal quality stream.

		:param bool ff:
			this input is not for the user please ignore.

		:rtype: :class:`Stream <Stream>`
		:returns: :class:`Stream <Stream>` object
		''' 
		if ff:
			av = int(len(self.data)/2)
			for n in range(av,av+2):
				if self.data[n].format == "webm":
					return self.data[n]
		else:return self.data[int(len(self.data)/2)]

	@property
	def ffmid(self) -> FFMPEG:
		'''
		Processing with mid quality video and audio

		:rtype: :class:`FFMPEG <FFMPEG>`
		:returns: :class:`FFMPEG <FFMPEG>` object
		'''
		vid = self.get_videos.mid(ff=True)
		aud = self.get_audios.mid(ff=True)
		return  FFMPEG([vid,aud])



	def resolution(self, quality) -> FFMPEG:
		'''
		Processing audio/video depend on resolution

		:rtype: :class:`FFMPEG <FFMPEG>`
		:returns: :class:`FFMPEG <FFMPEG>` object
		'''
		vid = ""
		for n in self.data:
			if quality == n.check("qualityLabel"):
				vid=n
				break
		if not vid:
			raise ResolutionError(f"quality {quality}p do not found")
		return  vid


	def ffresolution(self, quality:str) -> FFMPEG:
		'''
		Processing audio/video depend on resolution

		:param int quality:
			pass the quality type, Examples. ('144', '240', '360', '480', '720', '1080') more/less depend on your video.

		:rtype: :class:`FFMPEG <FFMPEG>`
		:returns: :class:`FFMPEG <FFMPEG>` object
		'''
		vid = self.get_videos.resolution(quality)
		aud = self.get_audios.mid(ff=True)
		return  FFMPEG([vid,aud])

	@property
	def all(self) -> List[Stream]:
		'''
		all will return the :class:`Stream <Stream>` objects in list

		:rtype: List of :class:`Stream <Stream>`
		:returns: list of :class:`Stream <Stream>` objects
		''' 
		return  self.All

	


# ddd=[{'itag': 18, 'mimeType': 'video/mp4', 'bitrate': 508303, 'width': 640, 'height': 360, 'lastModified': '1606480825357537', 'contentLength': '13744516', 'quality': 'medium', 'fps': 25, 'qualityLabel': '360p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 508202, 'audioQuality': 'AUDIO_QUALITY_LOW', 'approxDurationMs': '216363', 'audioSampleRate': '44100', 'audioChannels': 2, 'type': 'both', 'format': 'mp4', 'codecs': 'avc1.42001E,+mp4a.40.2', 's': '8Cu8CuLcpYZ5ct0QxiKT4YDqV9m8kQF2Uu1ECYyTBckkNBICsCJBqTOoAO84hAoWLhdgTq3obNuhubWZbygDpx9KefdgIARw8JQ0qOEON', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=18&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=oi6eTA7VyMX6HR8SoVyyoeIF&gir=yes&clen=13744516&ratebypass=yes&dur=216.363&lmt=1606480825357537&mt=1606580448&fvip=6&c=WEB&txp=2310222&n=1Imq9T7-tCc2gO0ZGV3&sparams=expire%2Cei%2Cip%2Cid%2Citag%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cratebypass%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhAKjaRpDh_jqk1D0FrgxCpmDunSeBih1uUTpB4pHUTwb9AiABquojbWbkUNYDO5z9w4v8kBCJgGS7pKV2Tes-BWpdoA%3D%3D&sig=AOq0QJ8wRAIgdfeK9xpDgybZWbuhuNbo3qTgdhLWoEh48OAoOTqBJCsCIBNkkcBTyYCN1uU2FQk8m9VqDY4TKixQ0tc5ZYpcLuC8'}, {'itag': 137, 'mimeType': 'video/mp4', 'bitrate': 4708327, 'width': 1920, 'height': 1080, 'initRange': {'start': '0', 'end': '741'}, 'indexRange': {'start': '742', 'end': '1301'}, 'lastModified': '1606480817048640', 'contentLength': '108097229', 'quality': 'hd1080', 'fps': 25, 'qualityLabel': '1080p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 3997678, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.640028', 's': '==w==wRfYq5cYZ6eInl4XT7hasdwhpmPlIqAhvtYlqNedoZ9DQICcr6VBe7sgVAC3AfWN-ZlgMxBYkAbR8x9rEL2laKK8TWgIQRw8JQ0qOYOF', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=137&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=108097229&dur=216.320&lmt=1606480817048640&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAJQvo9xFpPfXaaNO7MNXblBcCetwoq9m_tc3OILD2NCwAiEA0lLIxtiNSW4vSzbP6-36igJnponC4d_8IMcfRvHuQH8%3D&sig=AOq0QJ8wRQIgWT8KKal2LEr9x8RbAkYBxMglZ-NWfY3CAVgs7eBV6rcCIQD9ZodeNqlFtvhAqIlPmphwdsah7TX4lnIe6ZYc5qYfRw=='}, {'itag': 248, 'mimeType': 'video/webm', 'bitrate': 2464319, 'width': 1920, 'height': 1080, 'initRange': {'start': '0', 'end': '219'}, 'indexRange': {'start': '220', 'end': '988'}, 'lastModified': '1606480805859703', 'contentLength': '57206066', 'quality': 'hd1080', 'fps': 25, 'qualityLabel': '1080p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 2115608, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '=w1=w1n-_21KSD52DuHfRWceQDbsXH_tDlZmhlK1_fNNetnAEiAimv4tc-ny8K1cfAxJDjDVlot8SRIwOp2VRbSNoS9tsOAhIgRw8JQ0qO1Ox', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=248&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=57206066&dur=216.320&lmt=1606480805859703&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhALiAhu9BHsXzaTXooJOHoDloXpULEPUuh2ATp6N0cuWBAiEAwv06EWqRNFtvGGZfdoJa6Gr1srGqPuwF6fD-ykdqpzM%3D&sig=AOq0QJ8wRgIhAOst9SoNSbRV2pOwIRS8tolVDjDJx1fc1K8yn-ct4vmiAiEAnteNNf_xKlhmZlDt_HXsbDQecWRfHuD25DSK12_-n1w='}, {'itag': 136, 'mimeType': 'video/mp4', 'bitrate': 2253959, 'width': 1280, 'height': 720, 'initRange': {'start': '0', 'end': '739'}, 'indexRange': {'start': '740', 'end': '1299'}, 'lastModified': '1606480817047709', 'contentLength': '50981345', 'quality': 'hd720', 'fps': 25, 'qualityLabel': '720p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 1885404, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.64001f', 's': '==w==wsWVWkrnCLTaY4R5QTST2S8hoqgAOIOfYYb-mTRZRd_DQICAkpDeU0-TbxL1AcnV72mW_gfYv-AaqAF8oh5jI2bCGPgIQRw8JQ0qObOm', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=136&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=50981345&dur=216.320&lmt=1606480817047709&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhALk7gKc7Xud6LoTGXwHP8XpiFqBXF3JIS0g8aDj7FyimAiAKLKvdIbmXJAWRo9IGwnOlNnUh8dtaBqEyIko2wMsxgg%3D%3D&sig=AOq0QJ8wRQIgPGCb2Ij5ho8FAqaA-vYfg_Wm27Vncb1LxbT-0UeDpkACIQD_dRZRTm-mYYfOIOAgqoh8S2TSTQ5R4YaTLCnrkWVWsw=='}, {'itag': 247, 'mimeType': 'video/webm', 'bitrate': 1426017, 'width': 1280, 'height': 720, 'initRange': {'start': '0', 'end': '219'}, 'indexRange': {'start': '220', 'end': '980'}, 'lastModified': '1606480805867616', 'contentLength': '33002454', 'quality': 'hd720', 'fps': 25, 'qualityLabel': '720p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 1220504, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '=gQ=gQxG0WmzJyzjaEV-MHW1wJ2vZNIPWi3nDzmlVL8odgpAEiAAxPbbfIgdcSUFqAMfSDNreRxNyxNh_t0filTDPNEVzOAhIgRw8JQ0qOlOB', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=247&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=33002454&dur=216.320&lmt=1606480805867616&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRAIgQ4usgaIAyV5S6sq_bOG-6RlBHBLCSfiHlriRevBa7aQCID9XKN2EYEZZ1uWVDpJ0VavSsq2jKF16HhSFp7UqC0Oa&sig=AOq0QJ8wRgIhAOzVENPDTlif0t_hNxyNxRerNDSfMlqFUScdgIfbbPxAAiEApgdo8LVBmzDn3iWPINZv2Jw1WHM-VEajzyJzmW0GxQg='}, {'itag': 135, 'mimeType': 'video/mp4', 'bitrate': 1255827, 'width': 854, 'height': 480, 'initRange': {'start': '0', 'end': '740'}, 'indexRange': {'start': '741', 'end': '1300'}, 'lastModified': '1606480817062923', 'contentLength': '25831877', 'quality': 'large', 'fps': 25, 'qualityLabel': '480p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 955320, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.4d401e', 's': 'QKKQKKssBbMPzDNW37XLsAeRsCj0NLh-fZte06k7LytKEHIC0fyea0d0Oi1qrAexiZZTk6EeLH5qVu1gPBpx4aCR6MWgIARw8JQ0qOeO_', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=135&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=25831877&dur=216.320&lmt=1606480817062923&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAJB0f89tIXzeZqy5S1U6LwF9RkOrY2xZeiMk-D4VaexBAiEAmT58109DweOj7i6kv6T6L8P-t3o8H64ZqieVvSaCGhs%3D&sig=AOq0QJ8wRAIgWM6RCa4xpBPg1uVq5HLeE6kTZZixeerq1iO0d0aeyf0CIHEKtyL7k60_tZf-hLN0jCsReAsLX73WNDzPMbBssKKQ'}, {'itag': 244, 'mimeType': 'video/webm', 'bitrate': 717111, 'width': 854, 'height': 480, 'initRange': {'start': '0', 'end': '219'}, 'indexRange': {'start': '220', 'end': '957'}, 'lastModified': '1606480805914016', 'contentLength': '16479361', 'quality': 'large', 'fps': 25, 'qualityLabel': '480p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 609443, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '==w==wvCsltuKey9noYJdL4KMaYZm_kbKNWdSCohWC8zIvwqDQICMSTaARMe03pSwAeVJk6ddwCUcGydjIOj9ghRKR0unQbgIQRw8JQ0qOhOq', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=244&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=16479361&dur=216.320&lmt=1606480805914016&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIgVsNWt8sUJbUS4WtOxIzl0nvXyHabgJEll-rIJ-YUz8ACIQCpCZfMhdMnJlayXIp--NU_4i2I4xBH4P7rzbbqR1BVfg%3D%3D&sig=AOq0QJ8wRQIgbQnu0RKRhg9jOIjdyGcUCwdd6kJVehwSp30eMRAaTSMCIQDqwvIz8CWqoCSdWNKbk_mZYaMK4LdJYon9yeKutlsCvw=='}, {'itag': 134, 'mimeType': 'video/mp4', 'bitrate': 594687, 'width': 640, 'height': 360, 'initRange': {'start': '0', 'end': '740'}, 'indexRange': {'start': '741', 'end': '1300'}, 'lastModified': '1606480817054720', 'contentLength': '12073634', 'quality': 'medium', 'fps': 25, 'qualityLabel': '360p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 446510, 'highReplication': True, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.4d401e', 's': '-7I-7Ir4h-3GjOrgTR4zZdxdGgbQuOg1dWVM0VCCFYQxpBICIPp7LF6fmhfVNAxh37lrsW4jGQveCXBIBsS6jm21JlfgIARw8JQ0qOMO_', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=134&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=12073634&dur=216.320&lmt=1606480817054720&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAJbsaDlTRQK2g3XGFXXH0DwI58hyk5Nvyovj3Zj5g-fAAiEA57jz1vSgrzPyzdlI4aSWbAmFkyGiIlYHTyRzsOLj78Y%3D&sig=AOq0QJ8wRAIgflJ12mj6SsBIBXCevQGj4Wsrl73hxMNVfhmf6FL7pPICIBpxQYFCCV0_VWd1gOuQbgGdxdZz4RTgrOjG3-h4rI7-'}, {'itag': 243, 'mimeType': 'video/webm', 'bitrate': 380122, 'width': 640, 'height': 360, 'initRange': {'start': '0', 'end': '219'}, 'indexRange': {'start': '220', 'end': '957'}, 'lastModified': '1606480805968597', 'contentLength': '8903757', 'quality': 'medium', 'fps': 25, 'qualityLabel': '360p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 329280, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '==g==gCercLOBTkU5OOImbjt2mtnk2iz_tHPkKmtfAwOX2iIAiAl003NoBI0ZHa8zAXon8LN-ePLXYw_9vImvWzyQ1MgEKAhIQRw8JQ0qOtOb', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=243&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=8903757&dur=216.320&lmt=1606480805968597&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIgQxDWJ0AscWe_ydeit8R1ECPLOrvb3gKEZjWPBp040UQCIQC4BYpx_xDR8txUEUXH2SGWoLxmbUBmr0CIn-tLe8681Q%3D%3D&sig=AOq0QJ8wRQIhAKEgM1QyzWvmIv9_wYXLPe-NL8noXtz8aHZ0IBoN300lAiAIi2XOwAfbmKkPHt_zi2kntm2tjbmIOO5UkTBOLcreCg=='}, {'itag': 133, 'mimeType': 'video/mp4', 'bitrate': 259406, 'width': 426, 'height': 240, 'initRange': {'start': '0', 'end': '739'}, 'indexRange': {'start': '740', 'end': '1299'}, 'lastModified': '1606480817053058', 'contentLength': '5742765', 'quality': 'small', 'fps': 25, 'qualityLabel': '240p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 212380, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.4d4015', 's': '=Qe=Qez_YZgmXKsfP7LAVNF4DSjvY0WG3-DrgWIfb4WYYrvAEiA-gf1hgzWNRPVR3AOfD7_uPSFzKTSu2nzUx9E2ARTHRJAhIgRw8JQ0qOfOB', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=133&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=5742765&dur=216.320&lmt=1606480817053058&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhAP6mFT4jeS8cJCEdg3Ru-iOclN49bULRGP3LyYB2QG-1AiALPFet8Y44sFYLQVyIHbfTmSEI5CUEwdygBFdpjSevlA%3D%3D&sig=AOq0QJ8wRgIhAJRHTRA2E9xUzn2uSTKzFSPu_7DfOf3RVPRNWzgh1fg-AiEAvrYYW4bBIWgrD-3GW0YvjSD4FNVAL7PfsKXmgZY_zeQ='}, {'itag': 242, 'mimeType': 'video/webm', 'bitrate': 203358, 'width': 426, 'height': 240, 'initRange': {'start': '0', 'end': '218'}, 'indexRange': {'start': '219', 'end': '956'}, 'lastModified': '1606480805856031', 'contentLength': '4775218', 'quality': 'small', 'fps': 25, 'qualityLabel': '240p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 176598, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '==A==A5fK-T6vVEohgEZi2-iY0U1jdBS1IeRyLGJ4fAYWjLlDQIC86sAf2rpnAR05AOxamYqlbPn59j771UO2z4cy5JLgeNgIQRw8JQ0qOJOZ', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=242&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=4775218&dur=216.320&lmt=1606480805856031&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhANDQIB2K36i1oXVdW6rvDljUOlNYUJSmsawFyyoppXOXAiBLMJbtpB0t2083lc1eKWBTOuc-hpTws63ij5Q5s3bJ4Q%3D%3D&sig=AOq0QJ8wRQIgNegLJ5yc4z2OU177j95nPblqYmaxOJ50RAnpr2fAs68CIQDlLjWYAf4ZGLyReI1SBdj1U0Yi-2iZEghoEVv6T-Kf5A=='}, {'itag': 160, 'mimeType': 'video/mp4', 'bitrate': 116016, 'width': 256, 'height': 144, 'initRange': {'start': '0', 'end': '738'}, 'indexRange': {'start': '739', 'end': '1298'}, 'lastModified': '1606480817072996', 'contentLength': '2414363', 'quality': 'tiny', 'fps': 25, 'qualityLabel': '144p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 89288, 'approxDurationMs': '216320', 'type': 'video', 'format': 'mp4', 'codecs': 'avc1.4d400c', 's': '=Qq=Qq-a21oI6toil4Kg4JIuqOTL3X5xlexX6q8Cnu60z06AEiA5Ir8E7Xx3RHQgeA55qBKeW7VAi8-bK7bRdMURLOJvFIAhIgRw8JQ0qOCOg', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=160&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=2414363&dur=216.320&lmt=1606480817072996&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAPuJFRShk-NBokCaEQfMQNYHLCU8xzvaRuVhOiigx2skAiEAgbGKdR6BfZ2YGlFx1MLNDh2_PI6AWLtTqpmheD92FG4%3D&sig=AOq0QJ8wRgIhAIFvJOLRUMdRb7Kb-8iAV7WeKBq55CegQHR3xX7E8rI5AiEA60z06ung8q6Xxelx5X3LTOquIJ4gK4liot6Io12a-qQ='}, {'itag': 278, 'mimeType': 'video/webm', 'bitrate': 93329, 'width': 256, 'height': 144, 'initRange': {'start': '0', 'end': '218'}, 'indexRange': {'start': '219', 'end': '955'}, 'lastModified': '1606480805842854', 'contentLength': '2265367', 'quality': 'tiny', 'fps': 25, 'qualityLabel': '144p', 'projectionType': 'RECTANGULAR', 'averageBitrate': 83778, 'colorInfo': {'primaries': 'COLOR_PRIMARIES_BT709', 'transferCharacteristics': 'COLOR_TRANSFER_CHARACTERISTICS_BT709', 'matrixCoefficients': 'COLOR_MATRIX_COEFFICIENTS_BT709'}, 'approxDurationMs': '216320', 'type': 'video', 'format': 'webm', 'codecs': 'vp9', 's': '=sB=sBG08NSG5TjGPOlQWhGCV0f4FL6mDchBAaDxIR3fPu1AEiABHWRS2Ti6ny6jWA_-XpE2gg-NZQeKJhe76jmuwRgQ5MAhIgRw8JQ0qOxOV', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=278&aitags=133%2C134%2C135%2C136%2C137%2C160%2C242%2C243%2C244%2C247%2C248%2C278&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=video%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=2265367&dur=216.320&lmt=1606480805842854&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2316222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Caitags%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRgIhAPj5ESuvIdewMGAAd1DzH3IC4h5_HjxhR7uLhI1NCDsVAiEAhRmqliL-uQrFIU5NC3eMJuoZxyVmmeP0Er5-Ave0tBU%3D&sig=AOq0QJ8wRgIhAM5QgRwumj67ehJKeQZN-gg2EpX-_xWj6yn6iT2SRWHBAiEA1uPf3RIVDaABhcDm6LF4f0VCGhWQlOPGjT5GSN80GBs='}, {'itag': 140, 'mimeType': 'audio/mp4', 'bitrate': 130540, 'initRange': {'start': '0', 'end': '631'}, 'indexRange': {'start': '632', 'end': '927'}, 'lastModified': '1606480816199826', 'contentLength': '3503048', 'quality': 'tiny', 'projectionType': 'RECTANGULAR', 'averageBitrate': 129524, 'highReplication': True, 'audioQuality': 'AUDIO_QUALITY_MEDIUM', 'approxDurationMs': '216363', 'audioSampleRate': '44100', 'audioChannels': 2, 'loudnessDb': 5.6000004, 'type': 'audio', 'format': 'mp4', 'codecs': 'mp4a.40.2', 's': 't4kt4kOcUnNQTNcCi032JLSQRbeaF9sJgF81_ha6-oOU-AICAFw34g-EFpdG4AIS_Ej6F3qIP6706OT2l-OrjcGI4ZLgIARw8JQ0qO1OZ', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=140&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=audio%2Fmp4&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=3503048&dur=216.363&lmt=1606480816199826&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2311222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Citag%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIgeBDhOu6PUA8h3lub9xVLCCY8n-MoCXIwOHcH_TjQUb0CIQD75E9pkSyZqCzArNz-8FuNmxqnZiNtvzrVU-iZA6UXaQ%3D%3D&sig=AOq0QJ8wRAIgLZ4IGcjrO-l2TO6076PIq3F6jE_SI14GdpFE-g43wFACIA-UOo-6ah_Z8FgJs9FaebRQSLJ230iCcNTQNnUcOk4t'}, {'itag': 249, 'mimeType': 'audio/webm', 'bitrate': 61069, 'initRange': {'start': '0', 'end': '258'}, 'indexRange': {'start': '259', 'end': '629'}, 'lastModified': '1606480801923738', 'contentLength': '1548903', 'quality': 'tiny', 'projectionType': 'RECTANGULAR', 'averageBitrate': 57276, 'audioQuality': 'AUDIO_QUALITY_LOW', 'approxDurationMs': '216341', 'audioSampleRate': '48000', 'audioChannels': 2, 'loudnessDb': 5.5900002, 'type': 'audio', 'format': 'webm', 'codecs': 'opus', 's': '==A==AC1mcvc5-1V5NoKZBiSEBMbSvihWIqO-VkNSvwlq6czCQICkjy8TG-tZFog6AM0wdJ-RIlmagZNwoN8C_wPk3lr7_IgIQRw8JQ0qONOP', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=249&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=audio%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=1548903&dur=216.341&lmt=1606480801923738&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2311222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Citag%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhAOiDDe27rmdUNNs-epLIbTEFoBG0mzX4c_lD-WVk1TYiAiAK9VkOz6CMmRYGGKb0wMKRZCfrEXQ1jpnHjW0bh2mpig%3D%3D&sig=AOq0QJ8wRQIgI_7rl3kPw_C8NowNZgamlIR-Jdw0MN6goFZt-GT8yjkCIQCzc6qlwvSPkV-OqIWhivSbMBESiBZKoN5V1-5cvcm1CA=='}, {'itag': 250, 'mimeType': 'audio/webm', 'bitrate': 79396, 'initRange': {'start': '0', 'end': '258'}, 'indexRange': {'start': '259', 'end': '629'}, 'lastModified': '1606480802443010', 'contentLength': '2023978', 'quality': 'tiny', 'projectionType': 'RECTANGULAR', 'averageBitrate': 74843, 'audioQuality': 'AUDIO_QUALITY_LOW', 'approxDurationMs': '216341', 'audioSampleRate': '48000', 'audioChannels': 2, 'loudnessDb': 5.5900002, 'type': 'audio', 'format': 'webm', 'codecs': 'opus', 's': '==Q==QAnkUwmFvcgFvBSzYUs5sjwu5iS86WgK2zUX9W5WeqYAiAzHsYKOl8eMhUzzA6Bg0RQ69BQ0RfMhpI1F6hLJle00PAhIQRw8JQ0qOUOf', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=250&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=audio%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=2023978&dur=216.341&lmt=1606480802443010&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2311222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Citag%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRAIgVM6eli5OhRaIwvcX1u777pf5dgLDmX2QdPIhBBKNw08CICQzymOPsCFvy0R7MCg5me3tWjzLdEJ4uoCb7cuIybBt&sig=AOq0QJ8wRQIhAP00elJLh6F1IphMfR0QB96QR0gB6UzzUhMe8lOKYsHzAiAYqeW5W9Xfz2KgW68Si5uwjs5sUYzSBvFgcvFmwUknAQ=='}, {'itag': 251, 'mimeType': 'audio/webm', 'bitrate': 151452, 'initRange': {'start': '0', 'end': '258'}, 'indexRange': {'start': '259', 'end': '629'}, 'lastModified': '1606480802588428', 'contentLength': '3903186', 'quality': 'tiny', 'projectionType': 'RECTANGULAR', 'averageBitrate': 144334, 'audioQuality': 'AUDIO_QUALITY_MEDIUM', 'approxDurationMs': '216341', 'audioSampleRate': '48000', 'audioChannels': 2, 'loudnessDb': 5.5900002, 'type': 'audio', 'format': 'webm', 'codecs': 'opus', 's': '=op=opNOFMMh9DAAlZNrmAEPyMd-vWMqBUcqnxU9Wdk6ltkAEiAoJTbpBktBRa-bTAMODo8R3Cx2hGknSzAP8QIuiozhPNAhIgRw8JQ0qO9Oo', 'sp': 'sig', 'url': 'https://r6---sn-ci5gup-civl.googlevideo.com/videoplayback?expire=1606602349&ei=DXrCX4HYNsiwvQT6spfADA&ip=171.60.177.128&id=o-AFM4ZIRW6CXAcAMmIo499CWyIp_1F0RrUIOirOhdWA9c&itag=251&source=youtube&requiressl=yes&mh=YA&mm=31%2C29&mn=sn-ci5gup-civl%2Csn-ci5gup-cvhk&ms=au%2Crdu&mv=m&mvi=6&pl=20&initcwndbps=696250&vprv=1&mime=audio%2Fwebm&ns=DNCmEBtXb1sJZ_fUaL0Hm4cF&gir=yes&clen=3903186&dur=216.341&lmt=1606480802588428&mt=1606580448&fvip=6&keepalive=yes&c=WEB&txp=2311222&n=gkVd_Bk54e6S10gEAhN&sparams=expire%2Cei%2Cip%2Cid%2Citag%2Csource%2Crequiressl%2Cvprv%2Cmime%2Cns%2Cgir%2Cclen%2Cdur%2Clmt&lsparams=mh%2Cmm%2Cmn%2Cms%2Cmv%2Cmvi%2Cpl%2Cinitcwndbps&lsig=AG3C_xAwRQIhAPmUI7Y8f0OG27FsqT8mfKotv5ugpW-4fYReKyeC-g2kAiABXHbVv7sespnudprx0M3G7NgusVQ08MQwMcKoBrZWkg%3D%3D&sig=AOq0QJ8wRgIhANPhzoiuIQ8PAzSnkGh2xC3R8oDOM9Tb-aRBtkBpbTJoAiEAktl6kdWoUxnqcUBqMWv-dMyPEAmrNZlAAD9hMMFONpo='}]
# dtt=[]
# for n in ddd:
# 	dtt.append(Stream(n,"testinga"))


# # # print(Queue(dtt).format("mp4").Last.Url)

# # # for n in Queue(dtt).get_videos.get_itag(160):
# # # 	print(n)


# print(Queue(dtt).ffmid)





# -*- coding: utf-8 -*-
"""
youtube.py is really lite python liberary with fast downloader
"""

__title__ = "youtube.py"
__version__ = "1.1.6"
__author__ = "Mayank Gupta"
__license__ = "MIT License"
__copyright__ = "Copyright 2020 YouTube.py (Mayank Gupta)"
__github__ = "https://github.com/youtube-py/youtube.py"
__issues__ = "https://github.com/youtube-py/youtube.py/issues"
__docs__ = "https://youtube-python.mayankfawkes.xyz/"




# -*- coding: utf-8 -*-
"""
Here we have main classes

"""
from typing import Union, Callable
from typing import List
from typing import Dict
from typing import Any
import requests, json
import logging, sys
from .Stream import Stream
from .Stream import Queue
from time import ctime
from .Connection import HTTP
import concurrent.futures
from .Add_ons import (
	Get_vid_info_url,
	Get_r_vid_info_url,
	Get_video_id,
	Decipher_signature,
	Get_playlist_id,
	Extract_from_web_json
	)
from .Extract import (
	Is_age_restricted,
	Get_js,
	Get_video_info,
	Req_get,
	Get_playlist_videos,
	Continuation,
	Search_results
	)


logg = logging.getLogger(__name__)


HIGH = 0
MID = 1
LOW = 2


class Obj():
	def __init__(self,code:str, proxy:str="",id:int=1, name:str="") -> None:
		self.code = code
		self.proxy = proxy
		self.id = id
		self.name = name
		if not id:
			self.id = id
		if not name:
			self.name = "None"
		if not proxy:
			self.proxy = "None"

			
	def __str__(self) -> str:
		return f"<Class {self.__class__.__name__} VID={self.code} PROXY={self.proxy} ID={self.id} NAME={self.name} {hex(id(self))}>"
	def __repr__(self) -> str:
		return f"<Class {self.__class__.__name__} VID={self.code} PROXY={self.proxy} ID={self.id} NAME={self.name} {hex(id(self))}>"

class Video(Obj):
	'''
	Construct a :class:`Video <Video>` object.

	:param str url: Pass youtube valid video url. 
		ie. https://www.youtube.com/watch?v=9NQqaKz7eyI
	:param str proxy: (optional) pass the proxy url
		example: 
			``socks5://admin:admin@127.0.0.1:9050``,
			``http://admin:admin@127.0.0.1:8000``,
			``https://127.0.0.1:8000``

	:param int id: (optional) Video class id just to identify easily.
	:param object id: (optional) Pass the HTTP object.
	:param int name: (optional) if you want specify name.

	:rtype: None
	:returns: None
	'''
	def __init__(self,url:str, proxy:str="", id:int=1, http:HTTP=None, name:str="") -> None:
		self.http = http
		if not http:
			self.http = HTTP.MakeObject(proxy=proxy)
		self.video_id=Get_video_id(url)
		super().__init__(self.video_id,proxy,id, name)
		logg.debug(f"Working on Video id <{self.video_id}>")
		self.video_watch_url:str=f"https://www.youtube.com/watch?v={self.video_id}"

		self.js_url, self.video_html=Get_js(self.http,self.video_id)

		#Videos containers 
		self.fmts:Dict[Any,Any] = {}
		self.age_restricted:bool = False

		#Raw data store
		self.js:str=""
		self.video_info_url:str=""
		self.video_info:Dict[Any, Any]={}
		self.issig:bool=False

		self.Prefetch()


	def Prefetch(self) -> None:
		'''
		This will extract and do most of the work

		'''
		self.age_restricted = Is_age_restricted(self.video_html)
		if self.age_restricted:
			self.video_info_url = Get_r_vid_info_url(self.video_id)
		else:
			self.video_info_url = Get_vid_info_url(self.video_id,self.video_watch_url)

		self.fmts , self.issig, self.video_info =  Get_video_info(self.http,self.video_info_url)
		self.js=Req_get(self.http,self.js_url)

		if self.issig:
			Decipher_signature(self.fmts,self.js)

	@property
	def get_dict(self):
		'''
		This will return the raw dict of video streams urls
		'''
		return self.fmts
	

	@property
	def streams(self) -> Queue:
		'''
		Will returns :class:`Queue <Queue>`
				
		:rtype: :class:`Queue <Queue>`
		:returns: Returns the Queue object
		'''
		tmp = sorted(self.fmts,key= lambda x: x["itag"])
		return Queue([Stream(n, self.title) for n in self.fmts])

	@property
	def videoId(self) -> str:
		'''
		This will return the video id.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['videoId']

	@property
	def title(self) -> str:
		'''
		This will return the video title.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['title'].replace("+"," ")


	@property
	def length(self) -> int:
		'''
		This will return the video time in second.

		:rtype: int
		:returns: int
		'''
		return int(self.video_info['videoDetails']['lengthSeconds'])

	@property
	def keywords(self) -> List[str]:
		'''
		This will return the video keywords.

		:rtype: List[str]
		:returns: List[str]
		'''
		return self.video_info['videoDetails']['keywords']

	@property
	def channel(self) -> str:
		'''
		This will return the video channel id.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['channelId']

	@property
	def description(self) -> str:
		'''
		This will return the video description.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['shortDescription'].replace("+", " ")


	@property
	def author(self) -> str:
		'''
		This will return the video author.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['author'].replace("+", " ")

	@property
	def views(self) -> int:
		'''
		This will return the total video views.

		:rtype: int
		:returns: int
		'''
		return int(self.video_info['videoDetails']['viewCount'])

	@property
	def thumbnail(self) -> str:
		'''
		This will return the video thumbnail.

		:rtype: str
		:returns: str
		'''
		return self.video_info['videoDetails']['thumbnail']['thumbnails'][-1]['url']




class PlayList:
	'''
	Construct a :class:`PlayList <PlayList>`

	:param str url: Pass youtube valid playlist url. 
		ie. https://www.youtube.com/watch?v=9NQqaKz7eyI
	:param str proxy: (optional) pass the proxy url
		example: 
			``socks5://admin:admin@127.0.0.1:9050``
			``http://admin:admin@127.0.0.1:8000``
			``https://127.0.0.1:8080``
	:param bool process: (optional) Process will trigger the function for making objects of
			 :class:`Video <Video>` with the list of playlist videos default (False), When (True)
			 it will take some extra time depend on your CPU and Internet speed. Process will use 10 workers 
			 to create objects fasts.
	'''
	def __init__(self,url:str, proxy:str="", process:bool=False) -> None:
		self.http = HTTP.MakeObject(proxy=proxy)
		if proxy:
			logg.debug(f'Starting with proxy: {proxy}')


		# Holds playlist id `list=`
		self.playlist_id = Get_playlist_id(url)
		
		self.playlist_url = f"https://www.youtube.com/playlist?list={self.playlist_id}"

		self.html_parsec_json = Get_playlist_videos(self.http,self.playlist_url)

		# Holds the objects of class Video
		self.obj_videos = []
		self.get_videos()
		if process:
			self.processing()
	
	@property
	def get_dict(self):
		'''
		This will return the list of :class:`Video <Video>` if available or will just return the dict of vid and name, 
		Availability of :class:`Video <Video>` objects is depend on your ``PlayList(url,process=False)``, process will process
		all videos to :class:`Video <Video>` objects.
		'''
		if self.obj_videos:
			return self.obj_videos
		return self.videos

	@property
	def get_object(self):
		'''
		This will return the list of :class:`Video <Video>` objects, even when you ``process=False``.
		'''
		if not self.obj_videos:
			self.processing()
		return self.obj_videos
	

	def processing(self):
		'''
		Processing will process all videos to :class:`Video <Video>` object
		'''
		func = lambda x,y: Video(url=f'https://youtu.be/{x["vid"]}',http=self.http,name=f'{x["name"]}',id=y)
		with  concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
			for n in executor.map(func,self.videos,range(len(self.videos))):
				logg.debug(f'CLASS playlist created video object {n}')
				self.obj_videos.append(n)
		logg.debug(f'All objects of :class:`Video` created')

	def get_videos(self):
		'''
		Here we will process all videos dict
		'''
		self.videos, token = Extract_from_web_json(self.html_parsec_json)
		while token:
			token = Continuation(self.http, self.videos, token)

		logg.debug(f'Playlist collected {len(self.videos)}')

	def downloadall(self,quality:int) -> None:
		'''
		This is to download all playlist videos, remember this will download all videos
		synchronizly. Thiw function will download one file (audio/video) you may not satisfied with the quality. 

		+---------+---------+
		| Quality | Value   |
		+=========+=========+
		| HIGH    |   0     |
		+---------+---------+
		| MID     |   1     |
		+---------+---------+
		| LOW     |  2 or n |
		+---------+---------+

		:param int quality: Pass the quality type
		'''
		if not self.obj_videos:self.processing()
		logg.debug(f'Download all with quality type {quality}')
		if not self.obj_videos:
			self.processing()
		if quality == 0:
			for n in self.obj_videos:
				print(f'Downloading {n.title}')
				n.streams.get_both.high().download()
		elif quality == 1:
			for n in self.obj_videos:
				print(f'Downloading {n.title}')
				n.streams.get_both.mid().download()
		else:
			for n in self.obj_videos:
				print(f'Downloading {n.title}')
				n.streams.get_both.low().download()


class Search:
	'''
	This will search your query on youtube and will return videos.

	:param str query:
		Pass the query for search.
	:param str country: (optional)
		Pass the ISO 2 country code.
	:param str proxy: (optional) pass the proxy url
		example: 
			``socks5://admin:admin@127.0.0.1:9050``
			``http://admin:admin@127.0.0.1:8000``
			``https://127.0.0.1:8080``
	:param bool process: (optional) Process will trigger the function for making objects of
		 :class:`Video <Video>` with the list of playlist videos default (False), When (True)
		 it will take some extra time depend on your CPU and Internet speed. Process will use 10 workers 
		 to create objects fasts.
	
	:rtype: None
	:returns: None

	'''

	def __init__(self,query:str, country:str="", proxy:str="", process:bool=False) -> None:
		self.http = HTTP.MakeObject(proxy=proxy)

		if proxy:
			logg.debug(f'Starting with proxy: {proxy}')
		self.query = query
		self.country = country
		self.videos = []
		self.obj_videos = []
		self.search_start()
		if process:
			self.processing()
	@property
	def get_dict(self):
		'''
		This will return the list of :class:`Video <Video>` if available or will just return the dict of vid and name, 
		Availability of :class:`Video <Video>` objects is depend on your ``Search(query,process=False)``, process will process
		all videos to :class:`Video <Video>` objects.
		'''
		if self.obj_videos:
			return self.obj_videos
		return self.videos

	@property
	def first(self):
		'''
		Will return the first search video.

		:rtype: :class:`Video <Video>`
		:returns: This will return :class:`Video <Video>` of first search video.
		'''
		if self.obj_videos:
			return self.obj_videos[0]
		return Video(url=f'https://youtu.be/{self.videos[0]["vid"]}',http=self.http,name=f'{self.videos[0]["name"]}')

	@property
	def get_object(self):
		'''
		This will return the list of :class:`Video <Video>` objects, even when you ``process=False``.
		'''
		if not self.obj_videos:
			self.processing()
		return self.obj_videos

	def processing(self):
		'''
		Processing will process all videos to :class:`Video <Video>` object
		'''
		func = lambda x,y: Video(url=f'https://youtu.be/{x["vid"]}',http=self.http,name=f'{x["name"]}',id=y)
		with  concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
			for n in executor.map(func,self.videos,range(len(self.videos))):
				logg.debug(f'CLASS playlist created video object {n}')
				self.obj_videos.append(n)
		logg.debug(f'All objects of :class:`Video` created')

	def search_start(self):
		status = Search_results(self.http, self.query, self.videos, self.country)
		if status:
			logg.debug(f'Search completed and found {len(self.videos)} videos')




# -*- coding: utf-8 -*-
"""
Here are some function which helps extracting info
offline infomation parsec/filter/sort
"""
from typing import Union, Callable
from typing import List, NoReturn
from typing import Dict
from typing import Any
import logging
from re import compile, findall
from urllib.parse import unquote
from .Cipher import Cipher
from .Config import (__issues__,__github__)
from .Error import (InvalidVideoUrl, ParsecError)


logg = logging.getLogger(__name__)


def Get_video_id(url:str) -> str:
	'''
	Thie will return this video ID '&v='
	
	Example..

	>>> Get_Video_id("https://www.youtube.com/watch?v=12345678")
	'12345678'
	'''
	fu:str=""
	if "youtube.com" in url:
		fu={n.split("=")[0]:n.split("=")[1] for n in Get_url(url).split("?")[1].split("&")}["v"]
	elif "youtu.be" in url:
		fu=url.split("/")[-1]
	else:
		fu=""

	if fu:
		logg.debug(f"We successfully extracted VID<{fu}> from <{url}>")
		return fu
	else:
		logg.info(f'We dont support this url <{url}> please enter valid youtube watch url or contact {__issues__}')
		raise InvalidVideoUrl(f'We dont support this url <{url}> please enter valid youtube watch url or contact {__issues__}')

def Get_url(url:str) -> str:
	'''
	This will decode the url
	
	Example..

	>>> Get_Url("https://www.youtube.com/watchv%3D9NQqaKz7eyI")
	'https://www.youtube.com/watchv=9NQqaKz7eyI'
	'''
	return unquote(url)

def Get_vid_info_url(vid:str,whaturl:str) -> str:
	'''
	This will return the valid url of get_video_info

	The url for Video information do not works on 18+
	Videos

	:param str vid: 
		Pass the video id.
	:parma str whaturl:
		Pass the video url.

	:rtype: str
	:returns: the data url
	'''
	params={
		"video_id":vid,
		"ps":"default",
		"eurl":whaturl,
		"hl":"en_US",
	}
	return "https://youtube.com/get_video_info?"+"&".join(f"{n}={m}" for n,m in params.items())
	
def Get_r_vid_info_url(vid:str) -> str:
	'''
	This will return the valid url of get_video_info

	The url for Video information do works on 18+ Videos

	:param str vid: 
		Pass the video id.

	:rtype: str
	:returns: the data url
	'''
	params={
		"video_id":vid,
		"eurl":f"https://youtube.googleapis.com/v/{vid}",
		"sts":"",
	}
	return "https://youtube.com/get_video_info?"+"&".join(f"{n}={m}" for n,m in params.items())

def Decipher_signature(data:Dict[Any, Any],js:str) -> None:
	'''
	Here we will add the signature to url <&sig>

	:param dict data:
		All possible streams. 
	:param str js:
		Raw base.js js.

	:rtype: None
	:returns: nothing
	'''
	logg.debug("Decipher signature... ")

	enc=Cipher(js)
	for i,n in enumerate(data):
		if "&sig" in n["url"]:
			logg.debug("Video url already have signature token")
			continue
		sig=enc.Get_signature(n["s"])
		data[i]["url"] = n["url"] + "&sig=" + sig

	logg.debug("successfully Decipher signature... ")


#Playlist functions
def Get_playlist_id(url:str) -> str:
	'''
	Pass url of playlist and we will return the ID

	:param str url:
		pass the playlist url

	:rtype: str
	:returns: The params [**list**] from url.
	'''
	return {n.split("=")[0]:n.split("=")[1] for n in Get_url(url).split("?")[1].split("&")}["list"]


def Extract_from_web_json(data:Dict[str,str]) -> Dict[str,str]:
	'''
	This will extract the data we will get from website.

	:param Dict[str,str] data:
		Dict data of videos.

	:rtupe: Dict[str,str]
	:returns: the video ids dict with name. ``{"vid": "","name",""}``
	'''
	vids = []
	token = None

	for n in data["contents"]["twoColumnBrowseResultsRenderer"]["tabs"][0]["tabRenderer"]["content"]["sectionListRenderer"]["contents"][0]["itemSectionRenderer"]["contents"][0]["playlistVideoListRenderer"]["contents"]:
		if "playlistVideoRenderer" in n.keys():
			tmp = dict()
			tmp["vid"] = n["playlistVideoRenderer"]["videoId"]
			tmp["name"] = n["playlistVideoRenderer"]["title"]["runs"][0]["text"]
			vids.append(tmp)
		elif "continuationItemRenderer" in n.keys():
			token = n["continuationItemRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
		else:
			raise ParsecError(f'Serious Bug please report here {__issues__}')

	return vids, token
